/**
 * Created by janis dähne.
 */

interface JisonErrorHash {
  /**
   * expected keywords
   */
  expected: string[]
  line: number
  loc: {
    first_column: number
    first_line: number
    last_column: number
    last_line: number
  }
  recoverable: boolean
  text: string
  token: string
}

interface TerminalsTable {
  [tokenNum: number]: any //TODO not sure about typing...
}

interface JisonParser {

  yy: {
    parseError: (err: string, hash: JisonErrorHash) => void
  }

  lexer: JisonLexer
  terminals_: any //TODO not sure about typing...

  parse: (input: string) => ClassDiagram

}

interface TokenPair {
  token: string
  text: string
}

interface JisonLexer {
  yytext: string
  setInput: (input: string) => void
  lex: () => number | string //TODO not sure about typing...
  done: boolean
}

class PumclCompiler {

  /**
   * the parser
   */
  parser: JisonParser
  /**
   * true: only log errors to console.error
   * false: throw error and stop
   *
   * to disable errors or to not log to console.error set
   * parser.yy.parseError to custom function
   * @type {boolean}
   */
  isErrorRecoveryEnabled: boolean = false

  constructor(parser: JisonParser, isErrorRecoveryEnabled: boolean = false) {
    this.parser = parser

    this.isErrorRecoveryEnabled = isErrorRecoveryEnabled

    this.parser.yy.parseError = (err, hash) => {

      if (this.isErrorRecoveryEnabled) {
        console.error(err)
      } else {
        throw err
      }
    }
  }

  private preProcessInput(program: string): string {

    //the grammar forces a new line at the end of the program...
    if (program[program.length - 1] != '\n') {
      program = program + '\n'
    }

    return program

  }

  /**
   * tokenizes the input
   *
   * CAN THROW
   * @param {string} program
   * @returns {TokenPair[]}
   */
  tokenize(program: string): TokenPair[] {
    program = this.preProcessInput(program)

    let tokens: TokenPair[] = []

    this.parser.lexer.setInput(program)
    let token: any = 0
    while (!this.parser.lexer.done) {
      token = this.parser.lexer.lex();
      if (token in this.parser.terminals_) {
        token = this.parser.terminals_[token]
      }
      tokens.push({token: token, text: this.parser.lexer.yytext})
    }

    return tokens
  }

  /**
   * parses the input and returns a post processed diagram
   *
   * CAN THROW
   * @param {string} program
   * @param doPostProcessing
   * @returns {ClassDiagram}
   */
  parse(program: string, doPostProcessing = true): ClassDiagram {
    program = this.preProcessInput(program)

    let result: ClassDiagram

    //if we catch here we can never throw (no error recovery)
    result = this.parser.parse(program)

    if (doPostProcessing) {
      this.postProcessResult(result)
    }

    return result
  }

  /**
   * some post process steps
   * @param {ClassDiagram} diagram
   */
  postProcessResult(diagram: ClassDiagram): void {

    //1. set (parent) packages & prefix names
    ModelHelper.setParentPackages(diagram)

    //2. check for duplicates
    //e.g. in a package class names must be unique...

    BasicValidation.validateUniqueness(diagram)

    //2. check if all types are defined
    //TODO check if all types are defined
    //also for generic definitions...

    //3.
    //merge same types e.g. we create a new type
    //every time we found it but it's the same type (but different instances...)
    //it would be nice if all references refer to the same object instance

  }
}


/**
 * a type map to store all identifiers (a class, interface, enum, note)
 */
interface IdentifiersMap {
  [identifier: string]: Class | Interface | Enumeration | AssociationClass | Package | Note | PointingNote | any | undefined
}

interface PrimitiveTypesMap {
  [typeName: string]: Type | any | undefined
}

class ModelHelper {
  /**
   * sets all parent packages and names (names inside packages) correctly
   * @param {ClassDiagram} model
   * @returns {ClassDiagram}
   */
  static setParentPackages(model: ClassDiagram): ClassDiagram {

    ModelHelper.setParentPackagesInPackage(model, null)

    return model
  }

  /**
   * sets all names and parent packages recursively
   * @param {PackageShaped} packageWithEntities
   * @param {Package} parentPackage
   * @param packageNameToUseWithoutDelimiter
   */
  static setParentPackagesInPackage(packageWithEntities: PackageShaped, parentPackage: Package | null = null, packageNameToUseWithoutDelimiter: string | null = null): void {


    const packagePrefix = packageNameToUseWithoutDelimiter === null
      ? ''
      : packageNameToUseWithoutDelimiter + Constants.PackageDelimiter;

    //--- set default package / first level package
    for (const entity of packageWithEntities.classes) {
      entity.displayName_pk = packagePrefix
        + ModelHelper.getFlatEntityName(entity, entity2 => entity2.displayName_pk)

      entity.parentPackage = parentPackage
    }

    for (const entity of packageWithEntities.interfaces) {
      entity.displayName_pk = packagePrefix
        + ModelHelper.getFlatEntityName(entity, entity2 => entity2.displayName_pk)
      entity.parentPackage = parentPackage
    }

    for (const entity of packageWithEntities.enums) {
      entity.displayName_pk = packagePrefix
        + ModelHelper.getFlatEntityName(entity, entity2 => entity2.displayName_pk)
      entity.parentPackage = parentPackage
    }

    for (const entity of packageWithEntities.associationClasses) {
      entity.displayName_pk = packagePrefix
        + ModelHelper.getFlatEntityName(entity, entity2 => entity2.displayName_pk)
      entity.parentPackage = parentPackage
    }


    for (const entity of packageWithEntities.subPackages) {
      entity.name_pk = packagePrefix
        + ModelHelper.getFlatEntityName(entity, entity2 => entity2.name_pk)
      entity.parentPackage = parentPackage
    }

    //now set sub packages
    for (const subPackage of packageWithEntities.subPackages) {
      ModelHelper.setParentPackagesInPackage(subPackage, subPackage, subPackage.name_pk)
    }

  }

  //TODO test
  static getFlatEntityName<T extends ThingInsideContainer>(entity: T, getName: (entity: T) => string): string {

    const name = getName(entity)
    const index = name.lastIndexOf(Constants.PackageDelimiter)
    if (index === -1) {
      return name
    }
    return name.substring(index + Constants.PackageDelimiter.length)
  }

  static identifierExists(identifierMap: IdentifiersMap, identifier: string) {
    return !!identifierMap[identifier]
  }

  static primitiveTypeExists(primitiveTypesMap: PrimitiveTypesMap, primitiveType: string) {
    return !!primitiveTypesMap[primitiveType]
  }

}


class EqualChecker {

  static equalPackages(package1: Package | null, package2: Package | null): boolean {

    if (package1 === null && package2 === null) {
      //only default packages are null
      return true
    }
    else if (package1 != null && package2 !== null) {
      //both have names

      //they can have same name but different nesting
      //e.g. x::y::z and z::y::x
      let package1Name = ModelHelper.getFullPackageName(package1)
      let package2Name = ModelHelper.getFullPackageName(package2)

      return package1Name === package2Name
    }

    return false
  }

  static equalClasses(class1: Class, class2: Class): boolean {

    if (EqualChecker.equalPackages(class1.parentPackage, class2.parentPackage)) {
      //in same package
      return class1.displayName_pk === class2.displayName_pk
    }

    return false
  }

  static equalInterfaces(interface1: Class, interface2: Class): boolean {
    if (EqualChecker.equalPackages(interface1.parentPackage, interface2.parentPackage)) {
      //in same package
      return interface1.displayName_pk === interface2.displayName_pk
    }

    return false
  }

  static equalEnums(enum1: Enumeration, enum2: Enumeration): boolean {
    if (EqualChecker.equalPackages(enum1.parentPackage, enum2.parentPackage)) {
      //in same package
      return enum1.displayName_pk === enum2.displayName_pk
    }

    return false
  }

  static equalTypes(type1: Type, type2: Type): boolean {

    return type1.name_pk === type2.name_pk
      && type1.arrayDimension === type2.arrayDimension
    //TODO
    //&& type1.appliedGenerics.length === type2.appliedGenerics.length //e.g. Tuple<int, string> !== Tuple<int, string, int> in java, c# these are equal types??

  }

  static equalNotes(note1: Note, note2: Note): boolean {
    return note1.displayName === note2.displayName
  }

  static equalPointingNotes(pointingNote1: PointingNote, pointingNote2: PointingNote): boolean {
    //pointing notes have no name --> always different
    return false
  }

  static equalRelations(relation1: Relation, relation2: Relation): boolean {
    return EqualChecker.equalRelationNodes(relation1.startNode, relation2.startNode)
      && EqualChecker.equalRelationNodes(relation1.endNode, relation2.endNode)
      && relation1.name === relation2.name
  }

  static equalRelationNodes(relationNode1: RelationNode, relationNode2: RelationNode): boolean {
    //entity could be everything
    return relationNode1.entityName === relationNode2.entityName //works because full qualified names
      && relationNode1.propertyName === relationNode2.propertyName
  }
}


class CommonHelper {
  /**
   * checks every item in the array against every other item in the array with the given predicate
   * @param {T[]} array
   * @param {(item1: T, item2: T) => boolean} predicate
   * @returns {boolean} true: two items satisfied the predicate, false: all items are unique
   */
  static matchAny<T>(array: T[], predicate: (item1: T, item2: T) => boolean): boolean {

    for (let i = 0; i < array.length; i++) {
      let item1 = array[i];


      for (let j = i + 1; j < array.length; j++) {
        let item2 = array[j];

        if (predicate(item1, item2)) {
          return false
        }

      }
    }

    return true

  }

}


/*
 a basic validator to check if
 all names are unique...

 for the model
 all packages needs to have different names (even if nested)
 all primitive types must be unique for definition of unique see equal methods


for a package all names for classes + interfaces + enums must be unique
 this also applies for the default package (not a real package)



 for a class all method property names must be unique
 for an enum all value names must be unique
 for an enum all value values must be unique

 for a class all methods need a different signature
 for a interface all methods need a different signature

 for classes, interfaces defined, applied generics needs to be unique (the name)

 ----------

 also check that we have the types where we only got the type names
 e.g.
 RelationNode
 all types
   - generics definitions


 then check if all identifiers are defined & used primitive types exists
 als check this for applied generics

 */


class BasicValidation {


  static validateUniqueness(model: ClassDiagram): boolean {

    //check uniqueness

    if (BasicValidation.areNotesUnique(model) === false) {
      return false
    }

    if (BasicValidation.arePrimitiveTypesUnique(model) === false) {
      return false
    }

    if (BasicValidation.arePointingNotesUnique(model) === false) {
      return false
    }

    if (BasicValidation.areRelationsUnique(model) === false) {
      return false
    }

    if (BasicValidation.areDefaultPackageClassesUnique(model) === false) {
      return false
    }

    if (BasicValidation.areDefaultPackageInterfacesUnique(model) === false) {
      return false
    }

    if (BasicValidation.areDefaultPackageEnumsUnique(model) === false) {
      return false
    }

    //only check first level all nested ones have the parent package as prefix
    if (BasicValidation.arePackagesUnique(model) === false) {
      return false
    }

    if (BasicValidation.areInterfaceGenericsUnique(model.interfaces) === false) {
      return false
    }

    if (BasicValidation.areClassGenericsUnique(model) === false) {
      return false
    }

    if (BasicValidation.areEnumValuesUnique(model) === false) {
      return false
    }


    return true

  }

  static validate(model: ClassDiagram): boolean {



    // ----------- TODO check types

    // also check that we have the types (entities) for all identifiers
    // e.g.
    //   RelationNode
    // all types
    // - generics definitions

    if (BasicValidation.areAllTypesDefined(model) === false) {
      return false
    }


    return true
  }


  static areNotesUnique(model: ClassDiagram): boolean {

    return !CommonHelper.matchAny(model.notes, (item1, item2) => EqualChecker.equalNotes(item1, item2))
  }

  static arePointingNotesUnique(model: ClassDiagram): boolean {
    return !CommonHelper.matchAny(model.pointingNotes, (item1, item2) => EqualChecker.equalPointingNotes(item1, item2))
  }

  static arePrimitiveTypesUnique(model: ClassDiagram): boolean {

    return !CommonHelper.matchAny(model.primitiveTypes, (item1, item2) => EqualChecker.equalTypes(item1, item2))
  }

  static areRelationsUnique(model: ClassDiagram): boolean {
    return !CommonHelper.matchAny(model.relations, (item1, item2) => EqualChecker.equalRelations(item1, item2))
  }

  static areDefaultPackageClassesUnique(model: ClassDiagram): boolean {
    return !CommonHelper.matchAny(model.classes, (item1, item2) => EqualChecker.equalClasses(item1, item2))
  }

  static areDefaultPackageInterfacesUnique(model: ClassDiagram): boolean {
    return !CommonHelper.matchAny(model.interfaces, (item1, item2) => EqualChecker.equalInterfaces(item1, item2))
  }

  static areDefaultPackageEnumsUnique(model: ClassDiagram): boolean {
    return !CommonHelper.matchAny(model.enums, (item1, item2) => EqualChecker.equalEnums(item1, item2))
  }

  static arePackagesUnique(model: ClassDiagram): boolean {
    return !CommonHelper.matchAny(model.subPackages, (item1, item2) => EqualChecker.equalPackages(item1, item2))
  }

  static areInterfaceGenericsUnique(interfaces: Interface[]): boolean {

    //check if all interface level generics are unique &
    //checks if all interface method level generics are unique

    for (let i = 0; i < interfaces.length; i++) {
      let _interface = interfaces[i];

      if (CommonHelper.matchAny(_interface.genericsDefinitions,
          (item1, item2) => item1.parameterName === item2.parameterName)) {
        return false
      }

      //check methods

      for (let j = 0; j < _interface.methods.length; j++) {
        let method = _interface.methods[j];


        if (CommonHelper.matchAny(method.genericsDefinitions,
            (item1, item2) => item1.parameterName === item2.parameterName)) {
          return false
        }
      }
    }

    return true
  }

  static areClassGenericsUnique(classes: Class[]): boolean {

    //check if all interface level generics are unique &
    //checks if all interface method level generics are unique

    for (let i = 0; i < classes.length; i++) {
      let _class = classes[i];

      //e.g. List<T, T>
      if (CommonHelper.matchAny(_class.genericsDefinitions,
          (item1, item2) => item1.parameterName === item2.parameterName)) {
        return false
      }

      //check methods

      //e.g. List {
      // <T, T>foo()
      // }
      for (let j = 0; j < _class.methods.length; j++) {
        let method = _class.methods[j];


        if (CommonHelper.matchAny(method.genericsDefinitions,
            (item1, item2) => item1.parameterName === item2.parameterName)) {
          return false
        }
      }
    }

    return true
  }

  static areEnumValuesUnique(enums: Enumeration[]): boolean {

    for (let i = 0; i < enums.length; i++) {
      let _enum = enums[i];

      //equal names or equal values (but 2 null values is allowed)
      //TODO check
      if (CommonHelper.matchAny(_enum.values,
          (item1, item2) => item1.name === item2.name || (item1.value === item2.value && item1.value !== null)
        )) {
        return false
      }

    }
    return true
  }


  static areAllTypesDefined(model: ClassDiagram, identifierMap: IdentifiersMap, primitiveTypesMap: PrimitiveTypesMap): boolean {

    if (BasicValidation.areAllRelationIdentifiersDefined(model.relations, identifierMap) === false) {
      return false
    }

    if (BasicValidation.areAllGivenTypesDefined(model.primitiveTypes, identifierMap, primitiveTypesMap) === false) {
      return false
    }

  }

  static areAllRelationIdentifiersDefined(relations: Relation[], identifierMap: IdentifiersMap): boolean {

    for (let i = 0; i < relations.length; i++) {
      let relation = relations[i];

      if (ModelHelper.identifierExists(identifierMap, relation.startNode.entityName) === false) return false

      if (ModelHelper.identifierExists(identifierMap, relation.endNode.entityName) === false) return false

    }
    return true
  }


  //also checks generics
  static areAllPrimitiveTypesDefined(primitiveTypes: Type[], identifierMap: IdentifiersMap, primitiveTypesMap: PrimitiveTypesMap): boolean {

    for (let i = 0; i < primitiveTypes.length; i++) {
      let primitiveType = primitiveTypes[i];

      //a type can also be a class name...

      let isPrimitiveType = ModelHelper.primitiveTypeExists(primitiveTypesMap, primitiveType.name_pk)
      let isComplexType = ModelHelper.identifierExists(identifierMap, primitiveType.name_pk)

      if (!isPrimitiveType && !isComplexType) {
        return false
      }

      //also check generics
      BasicValidation.areAllGivenTypesDefined(primitiveType.appliedGenerics, identifierMap, primitiveTypesMap)

    }

    return true
  }


  static areAllGivenGenericsDefinitionsDefined(genericsDefinitions: GenericsDefinition[], identifierMap: IdentifiersMap, primitiveTypesMap: PrimitiveTypesMap): boolean {

    for (let j = 0; j < genericsDefinitions.length; j++) {
      let type = genericsDefinitions[j];

      if (type.constraint == null) continue

      if (ModelHelper.primitiveTypeExists(primitiveTypesMap, type.constraint.name_pk) == false) {
        return false
      }

      if (ModelHelper.identifierExists(identifierMap, type.constraint.name_pk) == false) {
        return false
      }

      //if we have e.g. List<T: Item<int>> check Item generics too
      if (BasicValidation.areAllGivenTypesDefined(type.constraint.appliedGenerics, identifierMap,
          primitiveTypesMap) === false) {
        return false
      }
    }
    return true
  }

  static areAllClassTypesDefined(classes: Class[], identifierMap: IdentifiersMap, primitiveTypesMap: PrimitiveTypesMap): boolean {

    //check if all class level generics are unique &
    //checks if all class method level generics are unique
    //check if all properties

    for (let i = 0; i < classes.length; i++) {
      let _class = classes[i];

      //e.g. List<T : Item, T2>
      if (BasicValidation.areAllGivenGenericsDefinitionsDefined(_class.genericsDefinitions, identifierMap,
          primitiveTypesMap) === false) {
        return false
      }


      //check method generics
      for (let j = 0; j < _class.methods.length; j++) {
        let method = _class.methods[j];

        if (BasicValidation.areAllGivenGenericsDefinitionsDefined(method.genericsDefinitions, identifierMap,
            primitiveTypesMap) === false) {
          return false
        }
        method.returnType

        //return type
      }

    }
  }

  static isTypeDefined(): boolean {

  }

}