/**
 * Created by janis dähne.
 */

declare var Viz: any


interface Theme {

  fontSizeInPx: number | null
  fontName: string | null

  graphBgColor: string | null

  classBgColor: string | null
  classBorderColor: string | null
  classFontColor: string | null

  interfaceBgColor: string | null
  interfaceBorderColor: string | null
  interfaceFontColor: string | null

  enumBgColor: string | null
  enumBorderColor: string | null
  enumFontColor: string | null

  packageBgColor: string | null
  packageBorderColor: string | null
  packageFontColor: string | null

  noteBgColor: string | null
  noteBorderColor: string | null
  noteFrontColor: string | null

  relationFontColor: string | null

  relationLineColor: string | null
  drawEntityNamesBold: boolean | null
}

/**
 * colors taken from https://github.com/altercation/ethanschoonover.com/tree/master/projects/solarized#usage-development
 */
class SolarizedLightTheme implements Theme {

  fontSizeInPx = 12
  fontName = 'Bitstream Vera Sans'
  graphBgColor = '#fdf6e3'

  classBgColor = '#eee8d5'
  classBorderColor = '#93a1a1'
  classFontColor = '#839496'

  interfaceBgColor = '#d6f7d3'
  interfaceBorderColor = '#93a1a1'
  interfaceFontColor = '#839496'

  enumBgColor = '#eecfaa'
  enumBorderColor = '#93a1a1'
  enumFontColor = '#839496'

  packageBgColor = null
  packageBorderColor = '#93a1a1'
  packageFontColor = '#839496'

  noteBgColor = null
  noteBorderColor = '#93a1a1'
  noteFrontColor = '#839496'

  relationFontColor = '#839496'
  relationLineColor = '#839496'

  drawEntityNamesBold = true

}


class TransformToGraphviz {

  static dashedStyle: string | 'dashed' | 'dotted' = 'dashed'
  static edgeLabelSpacing = '   '
  static roundedPackages = false
  static theme: null | Theme = new SolarizedLightTheme()

  /**
   * if the note has no alias (no name) it is prefixed with this
   * and a number (starting from 0)
   * @type {string}
   */
  static anonymousNotePrefix = '_anonymous_note_'
  static anonymousNoteCounter = 0


  static readonly newLinSeparator = '\\l' //if we use n all is centered
  static readonly newLinSeparatorHtml = '<br align="left"/>' //if we use n all is centered
  static readonly newLinSeparatorHtmlCentered = '<br align="center"/>'
  static readonly realNewLine = '\\n'

  static readonly angelBracketOpenHtml = '&lt;'
  static readonly angelBracketCloseHtml = '&gt;'

  /**
   * https://stackoverflow.com/questions/7586376/graphviz-subgraph-doesnt-get-visualized
   * @type {string}
   */
  static readonly packageHelpPrefix = 'cluster'
  /**
   * true: divide links into two parts via an invisible node in between... but this gives us a small gap
   * false: normal edges
   * @type {boolean}
   */
  static readonly enableDividedLinks = true


  static readonly dividedRelationsIntermediateNodePrefix = 'r'


  static setTheme(theme: Theme): void {
    this.theme = theme
  }

  static generateCode(model: ClassDiagram): string {

    TransformToGraphviz.anonymousNoteCounter = 0

    //TODO lollipop interface  https://de.wikipedia.org/wiki/Schnittstelle_(UML)
    //TODO entity level generics better placed,
    //TODo pointing notes

    //see
    // http://www.ffnn.nl/pages/articles/media/uml-diagrams-using-graphviz-dot.php
    // https://fsteeg.wordpress.com/2006/11/17/uml-class-diagrams-with-graphviz/

    //console.log("start")


    let fontName = this.theme === null
      ? ''
      : `fontname="${this.theme.fontName}"`

    let fontSize = this.theme === null
      ? ''
      : `fontsize="${this.theme.fontSizeInPx}"`

    let graphBgColor =
      this.theme === null
        ? ``
        : `graph[bgcolor="${this.theme.graphBgColor}"]`


    let dotCode = `
    digraph ClassDiagram {
        forcelabels=true
        ${fontName}
        ${fontSize}
        ${graphBgColor}

        node [
          ${fontName}
          ${fontSize}
          shape = "record"
        ]

        edge [
          ${fontName}
          ${fontSize}
        ]
    `

    for (const _class of model.classes) {
      dotCode += ClassTransformer.generateClass(_class)
    }

    for (const _interface of model.interfaces) {
      dotCode += InterfaceTransformer.generateInterface(_interface)
    }

    for (const _enum of model.enums) {
      dotCode += EnumTransformer.generateEnum(_enum)
    }


    for (const _package of model.subPackages) {
      dotCode += PackageTransformer.generatePackage(_package)
    }

    for (const _note of model.notes) {
      dotCode += NoteTransformer.generateNote(_note)
    }

    for (const relation of model.relations) {
      dotCode += TransformToGraphviz.enableDividedLinks
        ? RelationTransformer.generateDividedRelation(relation)
        : RelationTransformer.generateRelation(relation)
    }

    for (const _associationClass of model.associationClasses) {
      dotCode += AssociationClassTransformer.generateAssociationClass(_associationClass)
    }

    return dotCode + '\n}'
  }


  static toImage(code: string,
                 outputFormat: 'svg' | 'png-image-element' | 'xdot' | 'plain' | 'ps' | 'json' = 'svg',
                 renderEngine: 'circo' | 'dot' | 'fdp' | 'neato' | 'osage' | 'twopi' = 'dot'): any {

    return Viz(code, {engine: renderEngine, format: outputFormat});
  }
}


class CommonTransformer {


  static encloseInAngelBrackets(s: string | null): string {
    if (s === null) return ''

    return `\\<${s}\\>`
  }

  static encloseInDoubleAngelBrackets(s: string | null): string {
    if (s === null) return ''
    return CommonTransformer.encloseInAngelBrackets(CommonTransformer.encloseInAngelBrackets(s))
  }

  static encloseInAngelBracketsHtml(s: string | null): string {
    if (s === null) return ''

    return `${TransformToGraphviz.angelBracketOpenHtml}${s}${TransformToGraphviz.angelBracketCloseHtml}`
  }

  static encloseInDoubleAngelBracketsHtml(s: string | null): string {
    if (s === null) return ''
    return CommonTransformer.encloseInAngelBracketsHtml(CommonTransformer.encloseInAngelBracketsHtml(s))
  }

  static transformVisibilityModifier(visibilityModifier: VisibilityModifier | null): string {

    if (visibilityModifier === null) return ''

    if (visibilityModifier === VisibilityModifier._public) {
      return '+'
    } else if (visibilityModifier === VisibilityModifier._private) {
      return '-'
    } else if (visibilityModifier === VisibilityModifier._protected) {
      return '#'
    } else if (visibilityModifier === VisibilityModifier._packagePrivate) {
      return '~'
    }

    return ''
  }

  static transformType(type: Type | null): string {

    if (type === null) return ''

    const arrayPart = ('[]' as any).repeat(type.arrayDimension ? type.arrayDimension : 0)

    let genericsPart = ''

    if (type.appliedGenerics.length > 0) {

      for (let i = 0; i < type.appliedGenerics.length; i++) {
        let appliedGenericType = type.appliedGenerics[i]

        if (i == 0) {
          genericsPart += `${CommonTransformer.transformType(appliedGenericType)}`
          continue
        }
        genericsPart += `, ${CommonTransformer.transformType(appliedGenericType)}`
      }

      genericsPart = CommonTransformer.encloseInAngelBracketsHtml(genericsPart)
    }

    return `${type.name_pk}${genericsPart}${arrayPart}`

  }

  /**
   * just returns something like <T, T2, ...>
   * @param {GenericsDefinition[]} genericsDefinitions
   * @param {(s: string) => string} encloseInBrackets
   * @returns {string}
   */
  static transformGenericsDefinitionList(genericsDefinitions: GenericsDefinition[], encloseInBrackets: (s: string | null) => string): string {

    if (genericsDefinitions.length === 0) return ''

    let code = ``
    for (let i = 0; i < genericsDefinitions.length; i++) {
      let genericsDefinition = genericsDefinitions[i];


      if (i === 0) {

        if (genericsDefinition.constraint) {
          code += `${genericsDefinition.parameterName} : ${CommonTransformer.transformType(
            genericsDefinition.constraint)}`
        }
        else {
          code += `${genericsDefinition.parameterName}`
        }
        continue
      }

      if (genericsDefinition.constraint) {
        code += `, ${genericsDefinition.parameterName} : ${CommonTransformer.transformType(
          genericsDefinition.constraint)}`
      }
      else {
        code += `, ${genericsDefinition.parameterName}`
      }

    }
    return encloseInBrackets(code)
  }


  /**
   * returns content for a xlabel
   * @returns {string}
   */
  static transformEntityLevelGenericDefinitions(genericsDefinitions: GenericsDefinition[]): string {

    if (genericsDefinitions.length === 0) return ''


    let code = `<table  style="dashed">`
    for (let i = 0; i < genericsDefinitions.length; i++) {

      let genericsDefinition = genericsDefinitions[i];

      code += `<tr><td border="0">`

      if (genericsDefinition.constraint) {
        code += `${genericsDefinition.parameterName} : ${CommonTransformer.transformType(
          genericsDefinition.constraint)}`
      }
      else {
        code += `${genericsDefinition.parameterName}`
      }

      code += `</td></tr>`

    }

    return `${code}</table>`

  }


}

class ClassTransformer {

  static generateClass(_class: Class): string {

    let propertiesCode = ClassTransformer.generateProperties(_class.properties)
    let methodsCode = ClassTransformer.generateMethods(_class.methods)

    let stereotype = CommonTransformer.encloseInDoubleAngelBracketsHtml(_class.stereotype)

    if (stereotype) {
      stereotype += `${TransformToGraphviz.newLinSeparatorHtmlCentered}`
    }

    // let genericDefinition = CommonTransformer.transformGenericsDefinitionList(_class.genericsDefinitions,
    //   CommonTransformer.encloseInDoubleAngelBracketsHtml)
    let genericDefinition = ''

    let displayName = _class.alias ? _class.alias : _class.displayName_pk


    let code = ''

    let entityGenericDefinition = CommonTransformer.transformEntityLevelGenericDefinitions(_class.genericsDefinitions)

    if (TransformToGraphviz.theme !== null) {

      if (TransformToGraphviz.theme.drawEntityNamesBold) {
        code = `
        "${_class.displayName_pk}"[
          label = <{${stereotype}<b>${displayName}</b> ${genericDefinition}|${propertiesCode}|${methodsCode}}>
          `
      } else {
        code = `
        "${_class.displayName_pk}"[
          label = <{${stereotype}${displayName}${genericDefinition}|${propertiesCode}|${methodsCode}}>
          `
      }

      if (TransformToGraphviz.theme.classBgColor !== null) {
        code += `, style=filled, fillcolor="${TransformToGraphviz.theme.classBgColor}"
        `
      }

      if (TransformToGraphviz.theme.classBorderColor !== null) {
        code += `, color="${TransformToGraphviz.theme.classBorderColor}"
        `
      }

      if (TransformToGraphviz.theme.classFontColor !== null) {
        code += `, fontcolor="${TransformToGraphviz.theme.classFontColor}"
        `
      }

      if (entityGenericDefinition) {
        code += `, xlabel=<${entityGenericDefinition}>`
      }


      code += `]`

    } else {
      code = `
        "${_class.displayName_pk}"[
          label = <{${stereotype}${displayName} ${genericDefinition}|${propertiesCode}|${methodsCode}}>
        ]`
    }


    return code
  }

  static generateProperties(properties: Property[]): string {

    let code = ``

    for (const property of properties) {

      let propCode = ''

      propCode += `${CommonTransformer.transformVisibilityModifier(property.visibilityModifier)}`


      propCode += ` ${property.name_pk}`

      if (property.type !== null) {
        propCode += ` : ${CommonTransformer.transformType(property.type)}`
      }

      if (property.otherModifier === OtherModifier._abstract) {
        propCode = `<i>${propCode}</i>`
      }
      else if (property.otherModifier === OtherModifier._static) {
        propCode = `<u>${propCode}</u>`
      }

      code += `${propCode}${TransformToGraphviz.newLinSeparatorHtml}`
    }

    return code
  }

  static generateMethods(methods: Method[]): string {

    let code = ''

    for (const method of methods) {

      let methodeCode = ''

      methodeCode += `${CommonTransformer.transformVisibilityModifier(method.visibilityModifier)}`


      let genericsDefinitions = CommonTransformer.transformGenericsDefinitionList(method.genericsDefinitions,
        CommonTransformer.encloseInAngelBracketsHtml)

      methodeCode += `${method.name_pk}${genericsDefinitions}()`

      if (method.returnType !== null) {
        methodeCode += ` : ${CommonTransformer.transformType(method.returnType)}`
      }

      if (method.otherModifier === OtherModifier._abstract) {
        methodeCode = `<i>${methodeCode}</i>`
      }
      else if (method.otherModifier === OtherModifier._static) {
        methodeCode = `<u>${methodeCode}</u>`
      }

      code += `${methodeCode}${TransformToGraphviz.newLinSeparatorHtml}`
    }

    return code

  }

}

class InterfaceTransformer {

  static generateInterface(_interface: Interface): string {

    let methodsCode = InterfaceTransformer.generateMethods(_interface.methods)

    let stereotypeText = 'interface'

    if (_interface.stereotype !== null) {
      stereotypeText += `, ${_interface.stereotype}`
    }

    let stereotype = CommonTransformer.encloseInDoubleAngelBracketsHtml(stereotypeText)

    let genericDefinition = CommonTransformer.transformGenericsDefinitionList(_interface.genericsDefinitions,
      CommonTransformer.encloseInAngelBracketsHtml)

    let displayName = _interface.alias ? _interface.alias : _interface.displayName_pk

    let code = ''
    if (TransformToGraphviz.theme !== null) {


      if (TransformToGraphviz.theme.drawEntityNamesBold) {
        code = `
          ${_interface.displayName_pk}[
            label = <{${stereotype}${TransformToGraphviz.newLinSeparatorHtmlCentered}<b>${displayName}</b> ${genericDefinition}|${methodsCode}}>
          `
      } else {
        code = `
          ${_interface.displayName_pk}[
            label = "{${stereotype}${TransformToGraphviz.newLinSeparatorHtmlCentered}${displayName} ${genericDefinition}|${methodsCode}}"
          `
      }

      if (TransformToGraphviz.theme.interfaceBgColor !== null) {
        code += `, style=filled, fillcolor="${TransformToGraphviz.theme.interfaceBgColor}"
        `
      }

      if (TransformToGraphviz.theme.interfaceBorderColor !== null) {
        code += `, color="${TransformToGraphviz.theme.interfaceBorderColor}"
        `
      }

      if (TransformToGraphviz.theme.interfaceFontColor !== null) {
        code += `, fontcolor="${TransformToGraphviz.theme.interfaceFontColor}"
        `
      }

      code += `]`


    } else {
      code = `
    "${_interface.displayName_pk}"[
      label = <{${stereotype}${TransformToGraphviz.newLinSeparatorHtmlCentered}${displayName} ${genericDefinition}|${methodsCode}}>
    ]
    `
    }


    return code
  }

  static generateMethods(methods: InterfaceMethod[]): string {

    let code = ''

    for (const method of methods) {

      let methodCode = ''

      let genericsDefinitions = CommonTransformer.transformGenericsDefinitionList(method.genericsDefinitions,
        CommonTransformer.encloseInAngelBracketsHtml)

      methodCode += `${method.name_pk}${genericsDefinitions}()`

      if (method.returnType !== null) {
        methodCode += ` : ${CommonTransformer.transformType(method.returnType)}`
      }

      code += `${methodCode}${TransformToGraphviz.newLinSeparatorHtml}`
    }

    return code

  }

}

class EnumTransformer {

  static generateEnum(_enum: Enumeration): string {

    let fields = EnumTransformer.generateEnumValue(_enum.values)
    let stereotypeText = 'enumeration'

    if (_enum.stereotype !== null) {
      stereotypeText += `, ${_enum.stereotype}`
    }

    let stereotype = CommonTransformer.encloseInDoubleAngelBracketsHtml(stereotypeText)

    let displayName = _enum.alias ? _enum.alias : _enum.displayName_pk

    let code = ''

    if (TransformToGraphviz.theme !== null) {

      if (TransformToGraphviz.theme.drawEntityNamesBold) {
        code = `
          "${_enum.displayName_pk}"[
              label = <{${stereotype}${TransformToGraphviz.newLinSeparatorHtmlCentered}<b>${displayName}</b>|${fields}}>
          `
      } else {
        code = `
          "${_enum.displayName_pk}"[
              label = "{${stereotype}${TransformToGraphviz.newLinSeparatorHtmlCentered}${displayName}|${fields}}"
          `
      }

      if (TransformToGraphviz.theme.enumBgColor !== null) {
        code += `, style=filled, fillcolor="${TransformToGraphviz.theme.enumBgColor}"
        `
      }

      if (TransformToGraphviz.theme.enumBorderColor !== null) {
        code += `, color="${TransformToGraphviz.theme.enumBorderColor}"
        `
      }

      if (TransformToGraphviz.theme.enumFontColor !== null) {
        code += `, fontcolor="${TransformToGraphviz.theme.enumFontColor}"
        `
      }

      code += `]`

    } else {
      code = `
        "${_enum.displayName_pk}"[
            label = <{${stereotype}${TransformToGraphviz.newLinSeparatorHtmlCentered}${displayName}|${fields}}>
        ]
        `
    }

    return code

  }

  static generateEnumValue(values: EnumValue[]): string {

    let code = ''

    for (const pair of values) {

      let valueCode = ''

      valueCode += `${pair.name}`

      if (pair.value !== null) {

        let valueText = ''

        if (typeof pair.value === 'string') {
          valueText = `"${pair.value}"`
        } else {
          valueText = `${pair.value}`
        }

        valueCode += ` = ${valueText}`
      }

      code += `${valueCode}${TransformToGraphviz.newLinSeparatorHtml}`
    }

    return code
  }

}

class RelationTransformer {

  static generateRelation(relation: Relation): string {

    //A --> B
    //in dot A is the tail, B is the head
    //startNode = tail
    //endNode = head
    let tailNode = relation.startNode
    let headNode = relation.endNode


    let relationCode = `"${relation.startNode.entityName}" -> "${relation.endNode.entityName}"`


    let relationLabelCode = ''

    if (relation.name !== null) {

      if (relation.relationDirection === RelationDirection.startToEnd) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}${relation.name} >"`

      } else if (relation.relationDirection === RelationDirection.endToStart) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}< ${relation.name}"`

      } else {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}${relation.name}"`
      }

    } else {

      if (relation.relationDirection === RelationDirection.startToEnd) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}>"`

      } else if (relation.relationDirection === RelationDirection.endToStart) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}<"`

      } else {
        relationLabelCode = ''
      }

    }

    //---start node options
    let startNodeLabel = ''

    if (relation.startNode.cardinality !== null) {
      startNodeLabel += `, taillabel="${TransformToGraphviz.edgeLabelSpacing}${relation.startNode.cardinality}`
    }

    if (relation.startNode.cardinality === null && relation.startNode.propertyName !== null) {
      startNodeLabel += `, taillabel="${TransformToGraphviz.edgeLabelSpacing}${CommonTransformer.transformVisibilityModifier(
        relation.startNode.visibilityModifier)} - ${relation.startNode.propertyName}"`
    }

    if (relation.startNode.cardinality !== null && relation.startNode.propertyName !== null) {
      startNodeLabel += `${CommonTransformer.transformVisibilityModifier(
        relation.startNode.visibilityModifier)} - ${relation.startNode.propertyName}"`
    }

    if (relation.startNode.cardinality !== null && relation.startNode.propertyName === null) {
      startNodeLabel += `"`
    }

    //---end node options
    let endNodeLabel = ''

    if (relation.endNode.cardinality !== null) {
      endNodeLabel += `, headlabel="${TransformToGraphviz.edgeLabelSpacing}${relation.endNode.cardinality}`
    }

    if (relation.endNode.cardinality === null && relation.endNode.propertyName !== null) {
      endNodeLabel += `, headlabel="${TransformToGraphviz.edgeLabelSpacing}${CommonTransformer.transformVisibilityModifier(
        relation.endNode.visibilityModifier)} - ${relation.endNode.propertyName}"`
    }

    if (relation.endNode.cardinality !== null && relation.endNode.propertyName !== null) {
      endNodeLabel += `${CommonTransformer.transformVisibilityModifier(
        relation.endNode.visibilityModifier)} - ${relation.endNode.propertyName}"`
    }

    if (relation.endNode.cardinality !== null && relation.endNode.propertyName === null) {
      endNodeLabel += `"`
    }

    let dashedCode = relation.isDashed ? `, style=${TransformToGraphviz.dashedStyle}` : ''

    let fontColor = ''
    let lineColor = ''

    if (TransformToGraphviz.theme !== null) {

      if (TransformToGraphviz.theme.relationFontColor != null) {
        fontColor = `, fontcolor="${TransformToGraphviz.theme.relationFontColor}"`
      }

      if (TransformToGraphviz.theme.relationLineColor != null) {
        lineColor = `, color="${TransformToGraphviz.theme.relationLineColor}"`
      }

    }

    let optionsCode =
      `[dir=both` +
      `, arrowtail=${RelationTransformer.generateRelationKindPair(relation.startNode.linkKind)}` +
      `, arrowhead=${RelationTransformer.generateRelationKindPair(relation.endNode.linkKind)}` +
      `${relationLabelCode}` +
      `${startNodeLabel}` +
      `${endNodeLabel}` +
      `${dashedCode}` +
      `${fontColor}` +
      `${lineColor}`
      + ']'


    return `
    ${relationCode} ${optionsCode}
    `

  }

  /**
   * generates two lines with an intermediate invisible edge in between
   * this is useful when we want to point to the invisible edge e.g. for association classes
   * @param {Relation} relation
   * @returns {string}
   */
  static generateDividedRelation(relation: Relation): string {

    //A --> B
    //in dot A is the tail, B is the head
    //startNode = tail
    //endNode = head

    let intermediateNodeName = TransformToGraphviz.dividedRelationsIntermediateNodePrefix + relation.relationId
    let relationPart1 = `"${relation.startNode.entityName}" -> "${intermediateNodeName}"`
    let relationPart2 = `"${intermediateNodeName}" -> "${relation.endNode.entityName}"`


    let relationLabelCode = ''

    if (relation.name !== null) {

      if (relation.relationDirection === RelationDirection.startToEnd) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}${relation.name} >"`

      } else if (relation.relationDirection === RelationDirection.endToStart) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}< ${relation.name}"`

      } else {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}${relation.name}"`
      }

    } else {

      if (relation.relationDirection === RelationDirection.startToEnd) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}>"`

      } else if (relation.relationDirection === RelationDirection.endToStart) {
        relationLabelCode = `, label="${TransformToGraphviz.edgeLabelSpacing}<"`

      } else {
        relationLabelCode = ''
      }

    }

    //---start node options
    let startNodeLabel = ''

    if (relation.startNode.cardinality !== null) {
      startNodeLabel += `, taillabel="${TransformToGraphviz.edgeLabelSpacing}${relation.startNode.cardinality}`
    }

    if (relation.startNode.cardinality === null && relation.startNode.propertyName !== null) {
      startNodeLabel += `, taillabel="${TransformToGraphviz.edgeLabelSpacing}${CommonTransformer.transformVisibilityModifier(
        relation.startNode.visibilityModifier)} - ${relation.startNode.propertyName}"`
    }

    if (relation.startNode.cardinality !== null && relation.startNode.propertyName !== null) {
      startNodeLabel += `${CommonTransformer.transformVisibilityModifier(
        relation.startNode.visibilityModifier)} - ${relation.startNode.propertyName}"`
    }

    if (relation.startNode.cardinality !== null && relation.startNode.propertyName === null) {
      startNodeLabel += `"`
    }

    //---end node options
    let endNodeLabel = ''

    if (relation.endNode.cardinality !== null) {
      endNodeLabel += `, headlabel="${relation.endNode.cardinality}`
    }

    if (relation.endNode.cardinality === null && relation.endNode.propertyName !== null) {
      endNodeLabel += `, headlabel="${TransformToGraphviz.edgeLabelSpacing}${CommonTransformer.transformVisibilityModifier(
        relation.endNode.visibilityModifier)} - ${relation.endNode.propertyName}"`
    }

    if (relation.endNode.cardinality !== null && relation.endNode.propertyName !== null) {
      endNodeLabel += `${TransformToGraphviz.edgeLabelSpacing}${CommonTransformer.transformVisibilityModifier(
        relation.endNode.visibilityModifier)} - ${relation.endNode.propertyName}"`
    }

    if (relation.endNode.cardinality !== null && relation.endNode.propertyName === null) {
      endNodeLabel += `"`
    }

    let dashedCode = relation.isDashed ? `, style=${TransformToGraphviz.dashedStyle}` : ''

    let fontColor = ''
    let lineColor = ''

    if (TransformToGraphviz.theme !== null) {

      if (TransformToGraphviz.theme.relationFontColor != null) {
        fontColor = `, fontcolor="${TransformToGraphviz.theme.relationFontColor}"`
      }

      if (TransformToGraphviz.theme.relationLineColor != null) {
        lineColor = `, color="${TransformToGraphviz.theme.relationLineColor}"`
      }

    }


    let optionsCodePart1 =
      `[dir=both` +
      `, arrowtail=${RelationTransformer.generateRelationKindPair(relation.startNode.linkKind)}` +
      `, arrowhead="none"` +
      `${relationLabelCode}` +
      `${startNodeLabel}` +
      //`${endNodeLabel}` +
      `${dashedCode}` +
      `${fontColor}` +
      `${lineColor}`
      + ']'

    let optionsCodePart2 =
      `[dir=both` +
      `, arrowtail="none"` +
      `, arrowhead=${RelationTransformer.generateRelationKindPair(relation.endNode.linkKind)}` +
      //`${relationLabelCode}` +
      //`${startNodeLabel}` +
      `${endNodeLabel}` +
      `${dashedCode}` +
      `${fontColor}` +
      `${lineColor}`
      + ']'


    return `
    ${intermediateNodeName} [shape=point, width="0.0", height="0.0"]
    ${relationPart1} ${optionsCodePart1}
    ${relationPart2} ${optionsCodePart2}
    `

  }

  static generateRelationKindPair(linkKind: LinkKind): string {

    if (linkKind === LinkKind.normal) {
      return 'none'
    }
    else if (linkKind === LinkKind.arrow) {
      return 'normal'
    }
    else if (linkKind === LinkKind.generalization) {
      return 'onormal'
    }
    else if (linkKind === LinkKind.composition) {
      return 'diamond'
    }
    else if (linkKind === LinkKind.aggregation) {
      return 'odiamond'
    }
    else if (linkKind === LinkKind.box) {
      return 'obox'
    }
    else if (linkKind === LinkKind.cross) {
      return 'dot' //TODO...
    }
    else if (linkKind === LinkKind.trident) {
      return 'crow'
    }
    else if (linkKind === LinkKind.circlePlus) {
      return 'odot'
    }
    else if (linkKind === LinkKind.generalization2) {
      return 'onormal'
    }

    return 'none'
  }

}

class PackageTransformer {

  static generatePackage(_package: Package): string {

    let packageName = _package.name_pk

    let fontColor = ''
    let backgroundColor = ''
    let borderColor = ''

    if (TransformToGraphviz.theme !== null) {

      fontColor = TransformToGraphviz.theme.packageFontColor === null
        ? ''
        : `fontcolor="${TransformToGraphviz.theme.packageFontColor}"`

      backgroundColor = TransformToGraphviz.theme.packageBgColor === null
        ? ''
        : `bgcolor="${TransformToGraphviz.theme.packageBgColor}"`

      borderColor = TransformToGraphviz.theme.packageBorderColor === null
        ? ''
        : `color="${TransformToGraphviz.theme.packageBorderColor}"`
    }

    let code = `
      subgraph "${TransformToGraphviz.packageHelpPrefix}${packageName}" {
        graph[label=<${packageName}>
        ${fontColor}
        ${backgroundColor}
        ${borderColor}
        ${TransformToGraphviz.roundedPackages ? ',style="rounded"' : '' }
        ]
    `

    for (const _class of _package.classes) {
      code += ClassTransformer.generateClass(_class)
    }

    for (const _interface of _package.interfaces) {
      code += InterfaceTransformer.generateInterface(_interface)
    }

    for (const _enum of _package.enums) {
      code += EnumTransformer.generateEnum(_enum)
    }

    for (const _associationClass of _package.associationClasses) {
      code += AssociationClassTransformer.generateAssociationClass(_associationClass)
    }

    for (const _subPackage of _package.subPackages) {
      code += PackageTransformer.generatePackage(_subPackage)
    }

    code += `\n}`

    return code
  }

}

class NoteTransformer {
  static generateNote(_note: Note): string {

    let code = ''
    let noteName = ''

    if (_note.alias === null) {
      noteName = TransformToGraphviz.anonymousNotePrefix + TransformToGraphviz.anonymousNoteCounter
      TransformToGraphviz.anonymousNoteCounter++
    } else {
      noteName = _note.alias
    }

    code = `
      "${noteName}"[shape="note"
    `

    if (TransformToGraphviz.theme !== null) {
      if (TransformToGraphviz.theme.noteBgColor !== null) {
        code += `, style=filled, fillcolor="${TransformToGraphviz.theme.noteBgColor}"
        `
      }

      if (TransformToGraphviz.theme.noteBorderColor !== null) {
        code += `, color="${TransformToGraphviz.theme.noteBorderColor}"
        `
      }

      if (TransformToGraphviz.theme.noteFrontColor !== null) {
        code += `, fontcolor="${TransformToGraphviz.theme.noteFrontColor}"
        `
      }
    }

    code += `, label="${_note.text}"
      ]`

    if (_note.forRelationId !== null) {

      let intermediateNodeName = TransformToGraphviz.dividedRelationsIntermediateNodePrefix + _note.forRelationId

      let fontColor = ''
      let lineColor = ''

      if (TransformToGraphviz.theme !== null) {

        if (TransformToGraphviz.theme.relationFontColor != null) {
          fontColor = `, fontcolor="${TransformToGraphviz.theme.relationFontColor}"`
        }

        if (TransformToGraphviz.theme.relationLineColor != null) {
          lineColor = `, color="${TransformToGraphviz.theme.relationLineColor}"`
        }
      }

        code += `
          ${noteName} -> ${intermediateNodeName} [style="dashed", arrowhead="none"${fontColor}${lineColor}]
        `
    }


    return code
  }
}

class AssociationClassTransformer {
  static generateAssociationClass(_associationClass: AssociationClass): string {

    //relation start is always the association class
    //relation end is always the intermediate node

    // --> arrowhead = intermediate node
    // --> arrowtail = association class

    let intermediateNodeName = TransformToGraphviz.dividedRelationsIntermediateNodePrefix + _associationClass.relationId
    let relationCode = `"${_associationClass.displayName_pk}" -> "${intermediateNodeName}"`


    let dashedCode = `, style=${TransformToGraphviz.dashedStyle}`

    let fontColor = ''
    let lineColor = ''

    if (TransformToGraphviz.theme !== null) {

      if (TransformToGraphviz.theme.relationFontColor != null) {
        fontColor = `, fontcolor="${TransformToGraphviz.theme.relationFontColor}"`
      }

      if (TransformToGraphviz.theme.relationLineColor != null) {
        lineColor = `, color="${TransformToGraphviz.theme.relationLineColor}"`
      }
    }

    let optionsCode =
      `[arrowsize=0.4,
        dir=both` +
      `, arrowtail="none"` +
      `, arrowhead="dot"` +
      //`${endNodeLabel}` +
      `${dashedCode}` +
      `${fontColor}` +
      `${lineColor}`
      + ']'


    return `
    ${relationCode} ${optionsCode}
    `

  }
}